#ifndef GPIB_H
#define GPIB_H

int gpib_init(int fd);
int gpib_addr(int fd, int addr);

int keithley_init(int fd);
int keithley_current(int fd, double amps);
double keithley_read(int fd);
int keithley_off(int fd);

int keithley2002_init(int fd);
int keithley2002_init2(int fd);
int keithley2002_init_volts(int fd);
double keithley2002_read(int fd);
double keithley2002_read2(int fd, double* tx);
#endif

